"""
This is the do file. Here, we create the tiles in the createTiles function and then we execute their pipelines in the process function
"""

import dask
from dask.distributed import Lock
from . import tile
import pickle
import os


@dask.delayed
def process(pipeline, temp_dir=None):
    if temp_dir:
        with Lock(str(pipeline[1])):
            temp_file = temp_dir + '/' + str(pipeline[1]) + '.pickle'
        pipeline[0].execute()
        try:
            os.remove(temp_file)
        except FileNotFoundError:
            print('Trying to supress ' + temp_dir + temp_file + ' but cannot found this file.')
    else:
        pipeline[0].execute()


def createTiles(output_dir, json_pipeline, temp_dir=None, files=None, pipelines=None, dry_run=False):
    delayedPipelines = []
    if pipelines:
        for p in pipelines:
            delayedPipelines.append(dask.delayed(process)(p, temp_dir))
    else:
        tiles = []
        for file in files:
            tiles.append(tile.Tile(file, output_dir, json_pipeline))

        for t in tiles:
            pipeline = t.pipeline()
            if not dry_run:
                temp_file = temp_dir + '/' + str(pipeline[1]) + '.pickle'
                with open(temp_file, 'wb') as outfile:
                    pickle.dump(pipeline, outfile)
                delayedPipelines.append(dask.delayed(process)(pipeline, temp_dir))
            else:
                delayedPipelines.append(dask.delayed(process)(pipeline))

    return delayedPipelines



"""
This is the main file, it contains the process-pipelines function you call in command line
"""

import json
import sys
import click
import dask
from dask import config as cfg
from dask.distributed import LocalCluster, Client
from distributed.diagnostics import MemorySampler
from os import listdir
from . import do
from . import file_manager
from matplotlib import pyplot as plt


@click.group()
@click.version_option('0.7.8')
def main():
    """A simple parallelization tool for 3d point clouds treatment"""
    pass


@main.command()
@click.option('-c', '--config', required=True, type=click.Path(exists=True))
@click.option('-nw', '--n_workers', required=False, type=int, default=3)
@click.option('-tpw', '--threads_per_worker', required=False, type=int, default=1)
@click.option('-dr', '--dry_run', required=False, type=int)
@click.option('-d', '--diagnostic', is_flag=True, required=False)
def process_pipelines(**kwargs):
    """Processing pipelines on many points cloud in parallel"""
    with open(kwargs.get('config'), 'r') as c:
        config = json.load(c)

    output_dir = config.get('directories').get('output_dir')
    temp_dir = config.get('directories').get('temp_dir')
    if len(listdir(temp_dir)) != 0:
        click.echo('Something went wrong during previous execution, there is some temp files in your temp directory.\n Beginning of the execution\n')
        pipelines = file_manager.getSerializedPipelines(temp_dir)
        delayed = do.createTiles(output_dir=output_dir, temp_dir=temp_dir, json_pipeline=config.get('pipeline'), pipelines=pipelines)
    else:
        click.echo('Beginning of the execution\n')
        if not kwargs.get('dry_run'):
            files = file_manager.getFiles(config.get('directories').get('input_dir'))
            delayed = do.createTiles(output_dir=output_dir, json_pipeline=config.get('pipeline'), temp_dir=temp_dir, files=files)
        else:
            files = file_manager.getFiles(config.get('directories').get('input_dir'), nFiles=kwargs.get('dry_run'))
            delayed = do.createTiles(output_dir=output_dir, json_pipeline=config.get('pipeline'), files=files, dry_run=kwargs.get('dry_run'))

    timeout = input('After how long of inactivity do you want to kill your worker (timeout)\n')

    cfg.set({'interface': 'lo'})
    cfg.set({'distributed.scheduler.worker-ttl': None})
    cfg.set({'distributed.comm.timeouts.connect': timeout})
    cluster = LocalCluster(n_workers=kwargs.get('n_workers'), threads_per_worker=kwargs.get('threads_per_worker'))
    client = Client(cluster)
    ms = MemorySampler()
    click.echo('Parallelization started.\n')
    with ms.sample(label='execution', client=client):
        dask.compute(*delayed)

    if kwargs.get('diagnostic'):
        ms.plot()
        plt.savefig(config.get('directories').get('output_dir') + '/memory-usage.png')

    click.echo('Job just finished.\n')


if __name__ == "__main__":
    args = sys.argv
    if "--help" in args or len(args) == 1:
        print("pdal_parallelizer")
    main()



"""
This is the tile class. Each file represent a tile, and for each tile we assign a pipeline written by the user
"""

import sys
import pdal
import json
import os


class Tile:
    def __init__(self, filepath, output_dir, json_pipeline):
        self.filepath = filepath
        self.output_dir = output_dir
        self.json_pipeline = json_pipeline

    def pipeline(self):
        filename = os.path.basename(self.filepath).split('.')[0]
        name = 'temp__' + filename
        output_dir = self.output_dir
        output_filename = f'{output_dir}/{filename}.las'

        with open(self.json_pipeline, 'r') as pipeline:
            p = json.load(pipeline)
            reader = list(filter(lambda x: x['type'] == 'readers.las', p))
            writer = list(filter(lambda x: x['type'] == 'writers.las', p))

            if not reader:
                sys.exit("Please add a reader to your pipeline.")
            elif not writer:
                sys.exit("Please add a writer to your pipeline.")

            reader[0]['filename'] = self.filepath
            writer[0]['filename'] = output_filename

            p = pdal.Pipeline(json.dumps(p))

        return p, name

    def __str__(self):
        return f'{self.filepath}'



"""
In the file_manager, we get all the files that we need for the executions. We can get all the files from the input directory, or get all the serialized pipelines of the temp directory
"""

import os.path
import pickle
from os import listdir
from os.path import join


def getFiles(input_directory, nFiles=None):
    files = [join(input_directory, f) for f in listdir(input_directory)]
    if not nFiles:
        return files
    else:
        filesSize = [(join(input_directory, f), os.path.getsize(join(input_directory, f))) for f in files]
        filesSize.sort(key=lambda tup: tup[1], reverse=True)
        sortedFiles = [filesSize[i][0] for i in range(len(filesSize))]
        return sortedFiles[0:nFiles]


def getSerializedPipelines(temp_directory):
    pipelines = []
    for tmp in listdir(temp_directory):
        with open(join(temp_directory, tmp), 'rb') as p:
            pipelines.append(pickle.load(p))

    return pipelines
